"""The module provides an estimation process given the simulated data set and the
initialization file.
"""
from grmpy.check.auxiliary import read_data
from grmpy.check.check import (
    check_est_init_dict,
    check_par_init_file,
    check_presence_estimation_dataset,
)
from grmpy.estimate.estimate_par import par_fit
from grmpy.estimate.estimate_semipar import semipar_fit
from grmpy.read.read import check_append_constant, read


def fit(init_file, semipar=False):
    """This function estimates the MTE based on a parametric normal model
    or, alternatively, via the semiparametric method of
    local instrumental variables (LIV).

    Parameters
    ----------
    init_file: yaml
        Initialization file containing parameters for the estimation
        process.

    Returns
    ------
    rslt: dict
        Result dictionary containing
        - quantiles
        - mte
        - mte_x
        - mte_u
        - mte_min
        - mte_max
        - X
        - b1
        - b0
    """

    # Load the estimation file
    dict_ = read(init_file, semipar)

    # Perform some consistency checks given the user's request
    check_presence_estimation_dataset(dict_)
    check_est_init_dict(dict_)

    # Semiparametric LIV Model
    if semipar:
        # Distribute initialization information.
        data = read_data(dict_["ESTIMATION"]["file"])
        dict_, data = check_append_constant(init_file, dict_, data, semipar=True)

        rslt = semipar_fit(dict_, data)

    # Parametric Normal Model
    else:
        # Perform some extra checks
        check_par_init_file(dict_)

        # Distribute initialization information.
        data = read_data(dict_["ESTIMATION"]["file"])
        dict_, data = check_append_constant(init_file, dict_, data, semipar=False)

        rslt = par_fit(dict_, data)

    return rslt



"""The module provides the simulation process."""
import numpy as np

from grmpy.check.check import check_sim_init_dict
from grmpy.estimate.estimate_par import calculate_criteria, process_data, start_values
from grmpy.read.read import read_simulation
from grmpy.simulate.simulate_auxiliary import (
    print_info,
    simulate_covariates,
    simulate_outcomes,
    simulate_unobservables,
    write_output,
)


def simulate(init_file):
    """This function simulates a user-specified version of the generalized Roy model."""
    init_dict = read_simulation(init_file)

    # We perform some basic consistency checks regarding the user's request.
    check_sim_init_dict(init_dict)

    # Distribute information
    seed = init_dict["SIMULATION"]["seed"]

    # Set random seed to ensure recomputabiltiy
    np.random.seed(seed)

    # Simulate unobservables of the model
    U = simulate_unobservables(init_dict)

    # Simulate observables of the model
    X = simulate_covariates(init_dict)

    # Simulate endogeneous variables of the model
    df = simulate_outcomes(init_dict, X, U)

    # Write output file
    df = write_output(init_dict, df)

    # Calculate Criteria function value
    if not init_dict["DETERMINISTIC"]:
        D, X1, X0, Z1, Z0, Y1, Y0 = process_data(df, init_dict)
        x0 = start_values(init_dict, D, X1, X0, Z1, Z0, Y1, Y0, "init")
        init_dict["AUX"]["criteria_value"] = calculate_criteria(
            x0, X1, X0, Z1, Z0, Y1, Y0
        )

    # Print Log file
    print_info(init_dict, df)

    return df



"""The module provides basic auxiliary functions for the test modules."""
import glob
import os
import shlex

import numpy as np


def cleanup(options=None):
    """The function deletes package related output files."""
    fnames = glob.glob("*.grmpy.*")

    if options is None:
        for f in fnames:
            os.remove(f)
    elif options == "regression":
        for f in fnames:
            if f.startswith("regression"):
                pass
            else:
                os.remove(f)
    elif options == "init_file":
        for f in fnames:
            if f.startswith("test.grmpy"):
                pass
            else:
                os.remove(f)


def dict_transformation(dict_):
    varnames = []
    vartypes = {}
    for section in ["TREATED", "UNTREATED", "CHOICE"]:
        for variable in dict_[section]["order"]:
            if variable not in varnames:
                vartypes[variable] = dict_[section]["types"][
                    dict_[section]["order"].index(variable)
                ]
    for section in ["TREATED", "UNTREATED", "CHOICE", "DIST"]:
        dict_[section]["params"] = np.around(dict_[section].pop("all"), 4).tolist()
        dict_[section].pop("types", None)

    dict_["varnames"] = varnames

    for variable in vartypes:
        if isinstance(vartypes[variable], list):
            vartypes[variable][1] = float(np.around(vartypes[variable][1], 4))
    dict_["VARTYPES"] = vartypes
    return dict_


def read_desc(fname):
    """The function reads the descriptives output file and returns a dictionary that
    contains the relevant parameters for test6 in test_integration.py.
    """
    dict_ = {}
    with open(fname) as handle:
        for i, line in enumerate(handle):
            list_ = shlex.split(line)
            if 7 <= i < 10:
                if list_[0] in ["ALL", "TREATED", "UNTREATED"]:
                    dict_[list_[0]] = {}
                    dict_[list_[0]]["Number"] = list_[1:]
            elif 20 <= i < 23:
                if list_[0] == "Observed":
                    dict_["ALL"][list_[0] + " " + list_[1]] = list_[2:]
                else:
                    dict_["ALL"][list_[0] + " " + list_[1] + " " + list_[2]] = list_[3:]
            elif 29 <= i < 32:
                if list_[0] == "Observed":
                    dict_["TREATED"][list_[0] + " " + list_[1]] = list_[2:]
                else:
                    dict_["TREATED"][
                        list_[0] + " " + list_[1] + " " + list_[2]
                    ] = list_[3:]
            elif 38 <= i < 41:
                if list_[0] == "Observed":
                    dict_["UNTREATED"][list_[0] + " " + list_[1]] = list_[2:]
                else:
                    dict_["UNTREATED"][
                        list_[0] + " " + list_[1] + " " + list_[2]
                    ] = list_[3:]

        # Process the string in int and float values
        for key_ in dict_.keys():

            dict_[key_]["Number"] = [int(i) for i in dict_[key_]["Number"]]
            for subkey in [
                "Observed Sample",
                "Simulated Sample (finish)",
                "Simulated Sample (start)",
            ]:
                dict_[key_][subkey] = [
                    float(j) for j in dict_[key_][subkey] if j != "---"
                ]

    return dict_



"""This module provides auxiliary functions for the import process of the init file."""
import numpy as np


def create_attr_dict_est(init_dict, semipar=False, include_constant=False):
    """This function processes the imported initialization file so that it fulfills the
    requirements for the subsequent estimation process.
    """
    init_dict["AUX"] = {"init_values"}
    init_values = []

    if semipar is True:
        if include_constant is True:
            init_dict = add_constant(init_dict, semipar)
        else:
            pass

        init_dict = read_keys_semipar(init_dict, init_values)

    # semipar is False
    else:
        if include_constant is True:
            init_dict = add_constant(init_dict, semipar)
        else:
            pass

        init_dict = read_keys_par(init_dict, init_values)

    init_dict = provide_auxiliary_information(init_dict, init_values)

    return init_dict


def create_attr_dict_sim(init_dict):
    """This function processes the imported initialization file so that it fulfills the
    requirements for the following simulation and estimation process.
    """
    init_dict["AUX"] = {"init_values"}
    init_values = []

    init_dict = read_keys_par(init_dict, init_values)
    init_dict = provide_auxiliary_information(init_dict, init_values)

    return init_dict


def add_constant(init_dict, semipar=False):
    """The function checks if the user has provided a constant
    for the relevant subsections:
    ["TREATED", "UNTREATED", "CHOICE"] for the parametric, and
    ["CHOICE"] for the semiparamteric estimation, respectively.
    """

    if semipar is True:
        if "const" not in init_dict["CHOICE"]["order"]:
            init_dict["CHOICE"]["order"].insert(0, "const")
            init_dict["CHOICE"]["params"] = np.array([1.0])
        else:
            pass

    # semipar is False
    else:
        for key in ["TREATED", "UNTREATED", "CHOICE"]:
            if "const" not in init_dict[key]["order"]:
                init_dict[key]["order"].insert(0, "const")
                init_dict[key]["params"] = np.array([1.0])
            else:
                pass

    return init_dict


def read_keys_par(init_dict, init_values):
    """This function reads the information provided by the
    ["TREATED", "UNTREATED", "CHOICE", "DIST"] keys for
    the simulation and parametric estimation.
    """
    for key in ["TREATED", "UNTREATED", "CHOICE", "DIST"]:
        if "params" in init_dict[key].keys():
            init_dict[key]["params"] = np.array(init_dict[key]["params"])
            init_values += list(init_dict[key]["params"])
        else:
            init_values += [0.0] * len(init_dict[key]["order"])

    if np.all(init_dict["DIST"]["params"] == 0):
        init_dict["DETERMINISTIC"] = True
    else:
        init_dict["DETERMINISTIC"] = False

    return init_dict


def read_keys_semipar(init_dict, init_values):
    """This function reads the information provided by the
    ["TREATED", "UNTREATED", "CHOICE"] keys for
    semiparametric estimation.
    """
    for key in ["TREATED", "UNTREATED", "CHOICE"]:
        if "params" in init_dict[key].keys():
            init_dict[key]["params"] = np.array(init_dict[key]["params"])
            init_values += list(init_dict[key]["params"])
        else:
            init_values += [0.0] * len(init_dict[key]["order"])

    return init_dict


def provide_auxiliary_information(init_dict, init_values):
    """This function generates auxiliary information
    given the parameters in the initialization dictionary
    """
    num_covars = len(
        set(
            init_dict["TREATED"]["order"]
            + init_dict["UNTREATED"]["order"]
            + init_dict["CHOICE"]["order"]
        )
    )

    covar_label = []
    for section in ["TREATED", "UNTREATED", "CHOICE"]:
        covar_label += [i for i in init_dict[section]["order"] if i not in covar_label]

        # Generate the AUX section that include some additional auxiliary information
        init_dict["AUX"] = {
            "init_values": np.array(init_values),
            "num_covars_choice": len(init_dict["CHOICE"]["order"]),
            "num_covars_treated": len(init_dict["TREATED"]["order"]),
            "num_covars_untreated": len(init_dict["UNTREATED"]["order"]),
            "num_paras": len(init_values) + 1,
            "num_covars": num_covars,
            "labels": covar_label,
        }

    return init_dict



"""The module provides unit tests for different aspects of the simulation process."""
import numpy as np
import pandas as pd
import pytest
from statsmodels.tools.numdiff import approx_fprime_cs

from grmpy.check.auxiliary import read_data
from grmpy.estimate.estimate import fit
from grmpy.estimate.estimate_par import (
    adjust_output,
    backward_transformation,
    calculate_criteria,
    calculate_se,
    check_rslt_parameters,
    create_rslt_df,
    gradient_hessian,
    log_likelihood,
    minimizing_interface,
    process_data,
    process_output,
    start_value_adjustment,
    start_values,
)
from grmpy.grmpy_config import TEST_RESOURCES_DIR
from grmpy.read.read import read
from grmpy.simulate.simulate import simulate
from grmpy.simulate.simulate_auxiliary import (
    construct_covariance_matrix,
    mte_information,
)
from grmpy.test.auxiliary import cleanup
from grmpy.test.random_init import generate_random_dict, print_dict


def test1():
    """The first test tests whether the relationships in the simulated datasets are
    appropriate in a deterministic and an un-deterministic setting.
    """
    constr = {"DETERMINISTIC": True}
    for _ in range(10):
        generate_random_dict(constr)
        df = simulate("test.grmpy.yml")
        dict_ = read("test.grmpy.yml")
        x_treated = df[dict_["TREATED"]["order"]]
        y_treated = (
            pd.DataFrame.sum(dict_["TREATED"]["params"] * x_treated, axis=1) + df.U1
        )
        x_untreated = df[dict_["UNTREATED"]["order"]]
        y_untreated = (
            pd.DataFrame.sum(dict_["UNTREATED"]["params"] * x_untreated, axis=1) + df.U0
        )

        np.testing.assert_array_almost_equal(df.Y1, y_treated, decimal=5)
        np.testing.assert_array_almost_equal(df.Y0, y_untreated, decimal=5)
        np.testing.assert_array_equal(df.Y[df.D == 1], df.Y1[df.D == 1])
        np.testing.assert_array_equal(df.Y[df.D == 0], df.Y0[df.D == 0])


def test2():
    """The second test  checks whether the relationships hold if the coefficients are
    zero in different setups.
    """
    constr = {"DETERMINISTIC": True}

    for _ in range(10):
        for case in ["ALL", "TREATED", "UNTREATED", "CHOICE", "TREATED & UNTREATED"]:
            dict_ = generate_random_dict(constr)

            if case == "ALL":
                for section in ["TREATED", "UNTREATED", "CHOICE"]:
                    dict_[section]["params"] = np.array(
                        [0.0] * len(dict_[section]["params"])
                    )
            elif case == "TREATED & UNTREATED":
                for section in ["TREATED", "UNTREATED"]:
                    dict_[section]["params"] = np.array(
                        [0.0] * len(dict_[section]["params"])
                    )
            else:
                dict_[case]["params"] = np.array([0.0] * len(dict_[case]["params"]))

            print_dict(dict_)

            dict_ = read("test.grmpy.yml")
            df = simulate("test.grmpy.yml")
            x_treated = df[dict_["TREATED"]["order"]]
            x_untreated = df[dict_["UNTREATED"]["order"]]

            if case == "ALL":
                np.testing.assert_array_equal(df.Y1, df.U1)
                np.testing.assert_array_equal(df.Y0, df.U0)
            elif case == "TREATED & UNTREATED":
                np.testing.assert_array_equal(df.Y1, df.U1)
                np.testing.assert_array_equal(df.Y0, df.U0)
                np.testing.assert_array_equal(df.Y[df.D == 1], df.U1[df.D == 1])
                np.testing.assert_array_equal(df.Y[df.D == 0], df.U0[df.D == 0])
            elif case == "TREATED":
                y_untreated = (
                    pd.DataFrame.sum(dict_["UNTREATED"]["params"] * x_untreated, axis=1)
                    + df.U0
                )
                np.testing.assert_array_almost_equal(df.Y0, y_untreated, decimal=5)
                np.testing.assert_array_equal(df.Y1, df.U1)

            elif case == "UNTREATED":
                y_treated = (
                    pd.DataFrame.sum(dict_["TREATED"]["params"] * x_treated, axis=1)
                    + df.U1
                )
                np.testing.assert_array_almost_equal(df.Y1, y_treated, decimal=5)
                np.testing.assert_array_equal(df.Y0, df.U0)
            else:
                y_treated = (
                    pd.DataFrame.sum(dict_["TREATED"]["params"] * x_treated, axis=1)
                    + df.U1
                )
                y_untreated = (
                    pd.DataFrame.sum(dict_["UNTREATED"]["params"] * x_untreated, axis=1)
                    + df.U0
                )
                np.testing.assert_array_almost_equal(df.Y1, y_treated, decimal=5)
                np.testing.assert_array_almost_equal(df.Y0, y_untreated, decimal=5)

            np.testing.assert_array_equal(df.Y[df.D == 1], df.Y1[df.D == 1])
            np.testing.assert_array_equal(df.Y[df.D == 0], df.Y0[df.D == 0])


def test3():
    """The fourth test checks whether the simulation process works if there are only
    treated or untreated Agents by setting the number of agents to one. Additionally the
    test checks if the start values for the estimation process are set to the init-
    ialization file values due to perfect separation.
    """
    constr = {"AGENTS": 1, "DETERMINISTIC": False}
    for _ in range(10):
        generate_random_dict(constr)
        dict_ = read("test.grmpy.yml")
        df = simulate("test.grmpy.yml")
        D, X1, X0, Z1, Z0, Y1, Y0 = process_data(df, dict_)
        start = start_values(dict_, D, X1, X0, Z1, Z0, Y1, Y0, "init")
        np.testing.assert_equal(dict_["AUX"]["init_values"][:(-6)], start[:(-4)])


def test4():
    """The fifth test tests the random init file generating process and the import
    process. It generates an random init file, imports it again and compares the entries
    in  both dictionaries.
    """
    for _ in range(10):
        gen_dict = generate_random_dict()
        init_file_name = gen_dict["SIMULATION"]["source"]
        print_dict(gen_dict, init_file_name)
        imp_dict = read(init_file_name + ".grmpy.yml")
        dicts = [gen_dict, imp_dict]
        for section in ["TREATED", "UNTREATED", "CHOICE", "DIST"]:
            np.testing.assert_array_almost_equal(
                gen_dict[section]["params"], imp_dict[section]["params"], decimal=4
            )
            if section in ["TREATED", "UNTREATED", "CHOICE"]:
                for dict_ in dicts:
                    if not dict_[section]["order"] == dict_[section]["order"]:
                        raise AssertionError()
                    if len(dict_[section]["order"]) != len(
                        set(dict_[section]["order"])
                    ):
                        raise AssertionError()
                    if dict_[section]["order"][0] != "X1":
                        raise AssertionError()

        for variable in gen_dict["VARTYPES"].keys():
            if variable not in imp_dict["VARTYPES"].keys():
                raise AssertionError()

            if gen_dict["VARTYPES"][variable] != imp_dict["VARTYPES"][variable]:
                raise AssertionError

        if gen_dict["VARTYPES"]["X1"] != "nonbinary":
            raise AssertionError

        for subkey in ["source", "agents", "seed"]:
            if not gen_dict["SIMULATION"][subkey] == imp_dict["SIMULATION"][subkey]:
                raise AssertionError()

        for subkey in [
            "agents",
            "file",
            "optimizer",
            "start",
            "maxiter",
            "dependent",
            "indicator",
            "comparison",
            "output_file",
        ]:
            if not gen_dict["ESTIMATION"][subkey] == imp_dict["ESTIMATION"][subkey]:
                raise AssertionError()


def test5():
    """The tests checks if the simulation process works even if the covariance between
    U1 and V and U0 and V is equal. Further the test ensures that the mte_information
    function returns the same value for each quantile.
    """
    for _ in range(10):
        generate_random_dict()
        init_dict = read("test.grmpy.yml")

        # We impose that the covariance between the random components of the potential
        # outcomes and the random component determining choice is identical.
        init_dict["DIST"]["params"][2] = init_dict["DIST"]["params"][4]

        # Distribute information
        coeffs_untreated = init_dict["UNTREATED"]["params"]
        coeffs_treated = init_dict["TREATED"]["params"]

        # Construct auxiliary information
        cov = construct_covariance_matrix(init_dict)

        df = simulate("test.grmpy.yml")

        x = df[
            list(set(init_dict["TREATED"]["order"] + init_dict["UNTREATED"]["order"]))
        ]

        q = [0.01] + list(np.arange(0.05, 1, 0.05)) + [0.99]
        mte = mte_information(coeffs_treated, coeffs_untreated, cov, q, x, init_dict)

        # We simply test that there is a single unique value for the marginal treatment
        #  effect.
        np.testing.assert_equal(len(set(mte)), 1)


def test6():
    """The test ensures that the transformation of the optimization value vector works
    appropriately. For this purpose the test creates some random values converts them,
    transformes the values back and compared the resulting values to the initial values.
    """
    for _ in range(100):
        cov = np.random.uniform(0, 1, 2)
        var = np.random.uniform(1, 2, 3)
        before = [var[0], cov[0] / var[0], var[2], cov[1] / var[2]]
        transformed = start_value_adjustment(
            [var[0], cov[0] / var[0], var[2], cov[1] / var[2]]
        )
        backward_transformed = backward_transformation(transformed)
        np.testing.assert_array_almost_equal(before, backward_transformed, decimal=6)


def test7():
    """We want to able to smoothly switch between generating and printing random
    initialization files.
    """
    for _ in range(10):
        generate_random_dict()
        dict_1 = read("test.grmpy.yml")
        print_dict(dict_1)
        dict_2 = read("test.grmpy.yml")
        np.testing.assert_equal(dict_1, dict_2)


def test8():
    """This test ensures that the random process handles the constraints dict
    appropriately if there the input dictionary is not complete.
    """
    for _ in range(10):
        constr = dict()
        constr["MAXITER"] = np.random.randint(0, 1000)
        constr["START"] = np.random.choice(["start", "init"])
        constr["AGENTS"] = np.random.randint(1, 1000)
        dict_ = generate_random_dict(constr)
        np.testing.assert_equal(constr["AGENTS"], dict_["SIMULATION"]["agents"])
        np.testing.assert_equal(constr["START"], dict_["ESTIMATION"]["start"])
        np.testing.assert_equal(constr["MAXITER"], dict_["ESTIMATION"]["maxiter"])


def test9():
    """This test checks if the start_values function returns the init file values if the
    start option is set to init.
    """
    for _ in range(10):
        constr = dict()
        constr["DETERMINISTIC"] = False
        generate_random_dict(constr)
        dict_ = read("test.grmpy.yml")
        true = []
        for key_ in ["TREATED", "UNTREATED", "CHOICE"]:
            true += list(dict_[key_]["params"])
        df = simulate("test.grmpy.yml")
        D, X1, X0, Z1, Z0, Y1, Y0 = process_data(df, dict_)
        x0 = start_values(dict_, D, X1, X0, Z1, Z0, Y1, Y0, "init")[:-4]
        np.testing.assert_array_equal(true, x0)


def test10():
    """This test checks if the refactor auxiliary function returns an unchanged init
    file if the maximum number of iterations is set to zero.
    """

    for _ in range(10):
        constr = dict()
        constr["DETERMINISTIC"], constr["AGENTS"] = False, 1000
        constr["MAXITER"], constr["START"], constr["OPTIMIZER"] = 0, "init", "BFGS"
        generate_random_dict(constr)
        dict_ = read("test.grmpy.yml")
        df = simulate("test.grmpy.yml")
        D, X1, X0, Z1, Z0, Y1, Y0 = process_data(df, dict_)
        start = start_values(dict_, D, X1, X0, Z1, Z0, Y1, Y0, "init")
        start = backward_transformation(start)

        rslt = fit("test.grmpy.yml")

        np.testing.assert_equal(start, rslt["opt_rslt"]["params"].values)


def test11():
    """This test ensures that the tutorial configuration works as intended."""
    fname = TEST_RESOURCES_DIR + "/tutorial.grmpy.yml"
    simulate(fname)
    fit(fname)


def test12():
    """This test checks if our data import process is able to handle .txt, .dta and .pkl
    files.
    """

    pkl = TEST_RESOURCES_DIR + "/data.grmpy.pkl"
    dta = TEST_RESOURCES_DIR + "/data.grmpy.dta"
    txt = TEST_RESOURCES_DIR + "/data.grmpy.txt"

    real_sum = -3211.20122
    real_column_values = [
        "Y",
        "D",
        "X1",
        "X2",
        "X3",
        "X5",
        "X4",
        "Y1",
        "Y0",
        "U1",
        "U0",
        "V",
    ]

    for data in [pkl, dta, txt]:
        df = read_data(data)
        sum_ = np.sum(df.sum())
        columns = list(df)
        np.testing.assert_array_almost_equal(sum_, real_sum, decimal=5)
        np.testing.assert_equal(columns, real_column_values)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test13():
    """This test checks if functions that affect the estimation output adjustment work as
    intended.
    """
    for _ in range(5):
        generate_random_dict({"DETERMINISTIC": False})
        df = simulate("test.grmpy.yml")
        init_dict = read("test.grmpy.yml")
        D, X1, X0, Z1, Z0, Y1, Y0 = process_data(df, init_dict)
        rslt_cont = create_rslt_df(init_dict)
        start = start_values(init_dict, D, X1, X0, Z1, Z0, Y1, Y0, "init")
        init_dict["AUX"]["criteria"] = calculate_criteria(start, X1, X0, Z1, Z0, Y1, Y0)
        init_dict["AUX"]["starting_values"] = backward_transformation(start)

        aux_dict1 = {"crit": {"1": 10}}

        x0, se = [np.nan] * len(start), [np.nan] * len(start)
        index = np.random.randint(0, len(x0) - 1)
        x0[index], se[index] = np.nan, np.nan

        x_processed, crit_processed, _ = process_output(
            init_dict, aux_dict1, x0, "notfinite"
        )

        np.testing.assert_equal(
            [x_processed, crit_processed],
            [init_dict["AUX"]["starting_values"], init_dict["AUX"]["criteria"]],
        )

        check1, flag1 = check_rslt_parameters(start, X1, X0, Z1, Z0, Y1, Y0, aux_dict1)
        check2, flag2 = check_rslt_parameters(x0, X1, X0, Z1, Z0, Y1, Y0, aux_dict1)

        np.testing.assert_equal([check1, flag1], [False, None])
        np.testing.assert_equal([check2, flag2], [True, "notfinite"])

        opt_rslt = {
            "x": start,
            "fun": 1.0,
            "success": 1,
            "status": 1,
            "message": "msg",
            "nit": 10000,
        }

        rslt = adjust_output(
            opt_rslt,
            init_dict,
            rslt_cont,
            start,
            "BFGS",
            "init",
            X1,
            X0,
            Z1,
            Z0,
            Y1,
            Y0,
            aux_dict1,
        )
        np.testing.assert_equal(rslt["opt_info"]["crit"], opt_rslt["fun"])
        np.testing.assert_equal(rslt["opt_info"]["warning"][0], "---")

        x_linalign = [0] * len(x0)
        (
            se,
            hess_inv,
            conf_interval_low,
            conf_interval_up,
            p_values,
            t_values,
            _,
        ) = calculate_se(x_linalign, 1, X1, X0, Z1, Z0, Y1, Y0)
        np.testing.assert_equal(se, [np.nan] * len(x0))
        np.testing.assert_equal(hess_inv, np.full((len(x0), len(x0)), np.nan))
        np.testing.assert_equal(conf_interval_low, [np.nan] * len(x0))
        np.testing.assert_equal(conf_interval_up, [np.nan] * len(x0))
        np.testing.assert_equal(t_values, [np.nan] * len(x0))
        np.testing.assert_equal(p_values, [np.nan] * len(x0))


def test14():
    """This test checks wether our gradient functions work properly."""
    constr = {"AGENTS": 10000, "DETERMINISTIC": False}

    for _ in range(10):

        generate_random_dict(constr)
        init_dict = read("test.grmpy.yml")
        print(init_dict["AUX"])
        df = simulate("test.grmpy.yml")
        D, X1, X0, Z1, Z0, Y1, Y0 = process_data(df, init_dict)
        num_treated = X1.shape[1]
        num_untreated = X1.shape[1] + X0.shape[1]

        x0 = start_values(init_dict, D, X1, X0, Z1, Z0, Y1, Y0, "init")
        x0_back = backward_transformation(x0)
        llh_gradient_approx = approx_fprime_cs(
            x0_back,
            log_likelihood,
            args=(X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, None, False),
        )
        llh_gradient = gradient_hessian(x0_back, X1, X0, Z1, Z0, Y1, Y0)
        min_inter_approx = approx_fprime_cs(
            x0,
            minimizing_interface,
            args=(X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, None, False),
        )
        _, min_inter_gradient = log_likelihood(
            x0_back, X1, X0, Z1, Z0, Y1, Y0, num_treated, num_untreated, None, True
        )
        np.testing.assert_array_almost_equal(
            min_inter_approx, min_inter_gradient, decimal=5
        )
        np.testing.assert_array_almost_equal(
            llh_gradient_approx, llh_gradient, decimal=5
        )

    cleanup()



"""This module provides the fixtures for the PYTEST runs."""
import os

import numpy as np
import pytest
import tempfile


@pytest.fixture(scope="module", autouse=True)
def set_seed():
    """Each test is executed with the same random seed."""
    np.random.seed(1223)


@pytest.fixture(scope="module", autouse=True)
def fresh_directory():
    """Each test is executed in a fresh directory."""
    os.chdir(tempfile.mkdtemp())



"""
This module contains the semiparametric estimation process.
"""
import kernreg as kr
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.api as sm
from skmisc.loess import loess

lowess = sm.nonparametric.lowess


def semipar_fit(dict_, data):
    """ "
    This function runs the semiparametric estimation of the
    marginal treatment effect via local instrumental variables.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].

    Returns
    ------
    rslt: dict
        Result dictionary containing
        - quantiles
        - mte
        - mte_x
        - mte_u
        - mte_min
        - mte_max
        - X
        - b1
        - b0
    """
    # Process information specified in the initialization file and
    # assign default values if missing
    bins, logit, bandwidth, gridsize, startgrid, endgrid = process_primary_inputs(dict_)
    trim, rbandwidth, reestimate_p, show_output = process_secondary_inputs(dict_)

    # Prepare the sample for the estimation process
    # Compute propensity score, define common support and trim the sample
    # data, prop_score = process_mte_data(
    #     dict_, data, logit, bins, trim, reestimate_p, show_output
    # )
    data = estimate_treatment_propensity(dict_, data, logit, show_output)

    X, Y, prop_score = trim_support(
        dict_, data, logit, bins, trim, reestimate_p, show_output
    )

    b0, b1_b0 = double_residual_reg(X, Y, prop_score)

    # # Construct the MTE
    # Generate the quantiles of the final MTE
    quantiles = np.linspace(startgrid, endgrid, gridsize)

    mte_x = mte_observed(X, b1_b0)
    mte_u = mte_unobserved_semipar(
        X, Y, b0, b1_b0, prop_score, bandwidth, gridsize, startgrid, endgrid
    )

    # Put the MTE together
    mte = mte_x.mean(axis=0) + mte_u

    # Account for variation in X
    mte_min = np.min(mte_x) + mte_u
    mte_max = np.max(mte_x) + mte_u

    rslt = {
        "quantiles": quantiles,
        "mte": mte,
        "mte_x": mte_x,
        "mte_u": mte_u,
        "mte_min": mte_min,
        "mte_max": mte_max,
        "X": X,
        "b0": b0,
        "b1": b1_b0 + b0,
    }

    return rslt


def process_primary_inputs(dict_):
    """
    This functions processes the parameters specified
    by the user in the initialization dictionary.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file).

    Returns
    -------
    bins: int
        Number of histogram bins used to determine common support.
    logit: bool
        Probability model for the choice equation.
        If True: logit, else: probit.
    bandwidth: float
        Kernel bandwidth for the local polynomial regression.
    gridsize: int
        Number of equally-spaced grid points of u_D over which the
        MTE shall be estimated.
    startgrid: int
        Start point of the grid of unobservable resistance (u_D),
        over which the MTE is evaluated.
    end_grind: int
        End point of the grid of unobservable resistance (u_D),
        over which the MTE is evaluated.
    """
    try:
        dict_["ESTIMATION"]["bins"]
    except KeyError:
        bins = 25
    else:
        bins = dict_["ESTIMATION"]["bins"]

    try:
        dict_["ESTIMATION"]["logit"]
    except KeyError:
        logit = True
    else:
        logit = dict_["ESTIMATION"]["logit"]

    try:
        dict_["ESTIMATION"]["bandwidth"]
    except KeyError:
        bandwidth = 0.32
    else:
        bandwidth = dict_["ESTIMATION"]["bandwidth"]

    try:
        dict_["ESTIMATION"]["gridsize"]
    except KeyError:
        gridsize = 500
    else:
        gridsize = dict_["ESTIMATION"]["gridsize"]

    try:
        dict_["ESTIMATION"]["ps_range"]
    except KeyError:
        prop_score_range = [0.005, 0.995]
    else:
        prop_score_range = dict_["ESTIMATION"]["ps_range"]

    start_grid = prop_score_range[0]
    endgrid = prop_score_range[1]

    return bins, logit, bandwidth, gridsize, start_grid, endgrid


def process_secondary_inputs(dict_):
    """
    This functions processes the secondary input parameters.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file).

    Returns
    -------
    trim: bool, default True
        Trim the data outside the common support, recommended.
    rbandwidth: float, default 0.05
        Bandwidth for the Double Residual Regression.
    reestimate_p: bool, default False
        Re-estimate P(Z) after trimming, not recommended.
    show_output: bool, default False
        Show intermediate outputs of the estimation process.
    """
    try:
        dict_["ESTIMATION"]["trim_support"]
    # Set default to True
    except KeyError:
        trim = True
    else:
        trim = dict_["ESTIMATION"]["trim_support"]

    try:
        dict_["ESTIMATION"]["reestimate_p"]
    # Set default to False
    except KeyError:
        reestimate_p = False
    else:
        reestimate_p = dict_["ESTIMATION"]["reestimate_p"]

    try:
        dict_["ESTIMATION"]["rbandwidth"]
    # Set default to 0.05
    except KeyError:
        rbandwidth = 0.05
    else:
        rbandwidth = dict_["ESTIMATION"]["rbandwidth"]

    try:
        dict_["ESTIMATION"]["show_output"]
    # Set default to True
    except KeyError:
        show_output = False
    else:
        show_output = dict_["ESTIMATION"]["show_output"]

    return trim, rbandwidth, reestimate_p, show_output


def estimate_treatment_propensity(dict_, data, logit, show_output=False):
    """
    This function estimates the propensity of selecting into treatment
    for both treated and untreated individuals based on instruments Z.
    Z subsumes all the observable components that influence the treatment
    decision, e.g. the decision to enroll into college (D = 1) or not (D = 0).

    Estimate propensity scores via Logit (default) or Probit.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].
    logit: bool
        Probability model for the choice equation.
        If True: logit, else: probit.
    show_output: bool
        If True, intermediate outputs of the estimation process are displayed.

    Returns
    -------
    data: pandas.DataFrame
        Propensity score (range between [0, 1]). Values closer to 1
        denote a higher inclination to treatment.
    """
    D = data[dict_["ESTIMATION"]["indicator"]].values
    Z = data[dict_["CHOICE"]["order"]]

    if logit is True:
        logitRslt = sm.Logit(D, Z).fit(disp=0)
        prop_score = logitRslt.predict(Z)

        if show_output is True:
            print(logitRslt.summary())

    else:
        probitRslt = sm.Probit(D, Z).fit(disp=0)
        prop_score = probitRslt.predict(Z)

        if show_output is True:
            print(probitRslt.summary())

    data.loc[:, "prop_score"] = prop_score

    return data


def trim_support(
    dict_, data, logit, bins=25, trim=True, reestimate_p=False, show_output=False
):
    """
    This function defines common support and trims the data.
    Optionally p is re-estimated on the trimmed sample.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file).
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].
    logit: bool
        Probability model for the choice equation.
        If True: logit, else: probit.
    bins: int
        Number of histogram bins used to determine common support.
    trim: bool
        Trim the data outside the common support.
    reestimate_p: bool
        Re-estimate P(Z) after trimming.
    show_output: bool
        If True, intermediate outputs of the estimation process are displayed.

    Returns
    -------
    X: pandas.DataFrame
        Trimmed X data (observables), i.e. observations outside the common support
        of the propensity score (prop_score) have been dropped.
        Sorted in ascending order by *prop_score*.
    Y: pandas.DataFrame
        Trimmed Y data (wage), i.e. observations outside the common support
        of the propensity score (prop_score) have been dropped.
        Sorted in ascending order by *prop_score*.
    prop_score: pandas.Series
        Propensity score (range between [0, 1]). Values closer to 1
        denote a higher inclination to treatment.
        Sorted in ascending order.
    """
    # Find common support
    prop_score = data["prop_score"]
    common_support = _define_common_support(dict_, data, bins, show_output)

    # Trim the data. Recommended.
    if trim is True:
        # data, prop_score = trim_data(prop_score, common_support, data)
        data = data[
            (data.prop_score >= common_support[0])
            & (data.prop_score <= common_support[1])
        ]
        prop_score = prop_score[
            (prop_score >= common_support[0]) & (prop_score <= common_support[1])
        ]

        # Optional. Not recommended
        # Re-estimate baseline propensity score on the trimmed sample
        if reestimate_p is True:
            # Re-estimate the parameters of the decision equation based
            # on the new trimmed data set
            data = estimate_treatment_propensity(dict_, data, logit, show_output)

        else:
            pass
    else:
        pass

    data = data.sort_values(by="prop_score", ascending=True)
    prop_score = prop_score.sort_values(axis=0, ascending=True)
    X = data[dict_["TREATED"]["order"]]
    Y = data[[dict_["ESTIMATION"]["dependent"]]]

    return X, Y, prop_score


def double_residual_reg(X, Y, prop_score, rbandwidth=0.05, show_output=False):
    """
    This function performs a Double Residual Regression (DDR)
    of X, Xp, and Y on *prop_score*.

    A local linear kernel regression (polynomial of degree 1)
    is implemented to generate the residuals.

    Parameters
    ----------
    X: pandas.DataFrame
        DataFrame of observables (i.e. covariates).
    Y: pandas.DataFrame
        Individuals' wage data.
    prop_score: pandas.Series
        Propensity score (range between [0, 1]). Values closer to 1
        denote a higher inclination to treatment.
        Sorted in ascending order.

    Returns
    -------
    b0: np.ndarray
        Beta0 coefficient of the DDR (no-intercept OLS regression
        of the residuals of X, Xp, and Y on *prop_score*).
    b1: np.ndarray
        Beta1 coefficient of the DDR (no-intercept OLS regression
        of the residuals of X, Xp, and Y on *prop_score*).
    """
    # 0) Construct Xp := X * P(z)
    Xp = _construct_Xp(X, prop_score)

    # 1) Fit a separate local linear regression of X, Xp, and Y on prop_score,
    # which yields residuals e_X, e_Xp, and e_Y.
    res_X = _generate_residuals(prop_score, X, rbandwidth)
    res_Xp = _generate_residuals(prop_score, Xp, rbandwidth)
    res_Y = _generate_residuals(prop_score, Y, rbandwidth)

    # Append res_X and res_Xp.
    col_names = list(X) + list(Xp)
    res_X_Xp = pd.DataFrame(np.append(res_X, res_Xp, axis=1), columns=col_names)

    # 2) Run a single OLS regression of e_Y on e_X and e_Xp without intercept:
    # e_Y = e_X * beta_0 + e_Xp * (beta_1 - beta_0),
    # to estimate the values of beta_0 and (beta_1 - beta_0).
    model = sm.OLS(res_Y, res_X_Xp)
    results = model.fit()
    b0 = results.params[: len(list(X))]
    b1_b0 = results.params[len(list(X)) :]

    if show_output is True:
        print(results.summary())

    return np.array(b0), np.array(b1_b0)


def mte_observed(X, b1_b0):
    """
    This function computes the observed component of the MTE (*mte_x*)
    that depends on observables X:

    mte = *mte_x* + mte_u

    Parameters
    ----------
    X: pandas.DataFrame
        Data of observables (covariates).
    b1_b0: np.ndarray
        Difference of the coefficients in the Double Residual Regression,
        i.e. the no-intercept OLS regression of the residuals of
        X, Xp, and Y on *prop_score*.

    Returns
    -------
    mte_x: np.ndarray
        Part of the MTE that depends on observables X.
    """
    mte_x = np.dot(X, b1_b0)

    return mte_x


def mte_unobserved_semipar(
    X, Y, b0, b1_b0, prop_score, bandwidth, gridsize, startgrid, endgrid
):
    """
    This function computes the unobserved component of the MTE
    in MTE = mte_x + *mte_u*, where *mte_u* depends on the unobserved
    esistance to treatment u_D.

    Parameters
    ----------
    X: pandas.DataFrame
        DataFrame of observables (i.e. covariates).
    Xp: pandas.DataFrame
        X data multiplied by *prop_score* X * P(z).
    Y: pandas.DataFrame
        Individuals' wage data.
    b0: np.ndarray
        Beta0 coefficient in the Double Residual Regression,
        i.e. the no-intercept OLS regression of the residuals of
        X, Xp, and Y on *prop_score*.
    b1_b0: np.ndarray
        Difference of the coefficients in the Double Residual Regression,
        i.e. the no-intercept OLS regression of the residuals of
        X, Xp, and Y on *prop_score*.
    prop_score: pandas.Series
        Propensity score (range between [0, 1]). Values closer to 1
        denote a higher inclination to treatment.
        Sorted in ascending order.
    bandwidth: float
        Kernel bandwidth for the local polynomial regression.
    gridsize: int
        Number of equally-spaced grid points of u_D over which the
        MTE shall be estimated.
    startgrid: int
        Start point of the grid of unobservable resistance (u_D),
        over which the MTE is evaluated.
    endgrid: int
        End point of the grid of unobservable resistance (u_D),
        over which the MTE is evaluated.

    Returns
    -------
    mte_u: np.ndarray
        Part of the MTE that depends on the unobserved resistance
        to treatment (u_D).
    """
    # 0) Construct Xp := X * P(z)
    Xp = _construct_Xp(X, prop_score)

    # Turn the X, Xp, and Y DataFrames as well as the
    # propensity score Series into np.ndarrays
    X_arr = np.array(X)
    Xp_arr = np.array(Xp)
    Y_arr = np.array(Y).ravel()
    prop_score = np.array(prop_score)

    # Compute the unobserved part of Y
    Y_tilde = Y_arr - np.dot(X_arr, b0) - np.dot(Xp_arr, b1_b0)

    # Estimate mte_u, the unobserved component of the MTE,
    # through a locally quadratic regression
    rslt_locpoly = kr.locpoly(
        x=prop_score,
        y=Y_tilde,
        derivative=1,
        degree=2,
        bandwidth=bandwidth,
        gridsize=gridsize,
        a=startgrid,
        b=endgrid,
    )

    mte_u = rslt_locpoly["curvest"]

    return mte_u


def _define_common_support(
    dict_,
    data,
    bins=25,
    show_output=False,
    figsize=(10, 6),
    fontsize=15,
    plot_title=False,
    save_output=False,
):
    """
    This function defines the common support as the region under the histograms
    where propensities in the treated and untreated subsample overlap.

    Carneiro et al (2011) choose 25 bins for a total sample of 1747
    observations, so *nbins*=25 is set as a default.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].
    bins: int
        Number of histogram bins used to determine common support.
    show_output: bool
        If True, intermediate outputs of the estimation process are displayed.
    figsize: tuple, default is (10, 6)
        Tuple denoting the size of the resulting figure.
    fontsize: int, default is 15
        Parameter specifying the font size used in the figure.
    plot_title: bool, default is False
        If True, a title for the figure is displayed.
    save_output: bool or str or PathLike or file-like object, default is False
        If False, the resulting plot is shown but not saved.
        If True, the MTE plot is saved as 'common_support.png'.
        Else, if a str or Pathlike or file-like object is specified,
        the plot is saved according to *save_output*.
        The output format is inferred from the extension ('png', 'pdf', 'svg'... etc.)
        By default, '.png' is assumed.

    Returns
    -------
    common_support: list
        List containing lower and upper bound of the propensity score.
    """
    hist, treated, untreated = _make_histogram(
        dict_, data, bins, show_output, figsize, fontsize, plot_title
    )
    lower_limit, upper_limit = _find_limits(hist, treated, untreated)
    common_support = [lower_limit, upper_limit]

    if show_output is True:
        print(
            """
    Common support lies beteen:

        {} and
        {}""".format(
                lower_limit, upper_limit
            )
        )

    if save_output is False:
        pass
    elif save_output is True:
        plt.savefig("common_support.png", dpi=300)
    else:
        plt.savefig(save_output, dpi=300)

    return common_support


def _make_histogram(
    dict_,
    data,
    bins=25,
    show_output=False,
    figsize=(10, 6),
    fontsize=15,
    plot_title=False,
):
    """
    This function plots the common supports, i.e. the overlapping regions under the
    histograms of both the treated and untreated individuals.
    The plot is only displayed if *show_output* is True.

    Parameters
    ----------
    dict_: dict
        Estimation dictionary. Returned by grmpy.read(init_file)).
    data: pandas.DataFrame
        Data set to perform the estimation on. Specified
        under dict_["ESTIMATION"]["file"].
    bins: int
        Number of histogram bins used to determine common support.
    show_output: bool
        If True, intermediate outputs of the estimation process are displayed.
    figsize: tuple, default is (10, 6)
        Tuple denoting the size of the resulting figure.
    fontsize: int, default is 15
        Parameter specifying the font size used in the figure.
    plot_title: bool, default is False
        If True, a title for the figure is displayed.

    Returns
    -------
    hist: np.ndarray
        Array containing positional parameters of the histogram bins.
    treated: list
        List containing the propensity scores of the treated
        individuals.
    untreated: list
        List containing the propensity scores of the treated
        individuals.
    """
    indicator = dict_["ESTIMATION"]["indicator"]

    treated = data[[indicator, "prop_score"]][data[indicator] == 1].values
    untreated = data[[indicator, "prop_score"]][data[indicator] == 0].values

    treated = treated[:, 1].tolist()
    untreated = untreated[:, 1].tolist()

    # Make the histogram using a list of lists
    fig = plt.figure(figsize=figsize)
    hist = plt.hist(
        [treated, untreated],
        bins=bins,
        weights=[
            np.ones(len(treated)) / len(treated),
            np.ones(len(untreated)) / len(untreated),
        ],
        density=0,
        alpha=0.55,
        label=["Treated", "Unreated"],
    )

    if show_output is True:
        plt.tick_params(axis="both", labelsize=14)
        plt.legend(loc="upper right", prop={"size": 14})
        plt.xticks(np.arange(0, 1.1, step=0.1))
        plt.grid(axis="y", alpha=0.25)
        plt.xlabel("$P$", fontsize=fontsize)
        plt.ylabel("$f(P)$", fontsize=fontsize)

        if plot_title is True:
            plt.title(r"Support of $P(\hat{Z})$ for $D=1$ and $D=0$")

    else:
        plt.close(fig)

    return hist, treated, untreated


def _find_limits(hist, treated, untreated):
    """
    Find the upper and lower limit of the common support.

    Parameters
    ----------
    hist: np.ndarray
        Array containing positional parameters of the histogram bins.
    treated: list
        List containing the propensity scores of the treated
        individuals.
    untreated: list
        List containing the propensity scores of the treated
        individuals.

    Returns
    -------
    lower_limit: float
        Lower limit of the common support.
    upper_limit: float
        Upper limit of the common support.
    """
    # Treated Sample
    # Set the lowest frequency observed in the treated subsample
    # as the default for the lower limit of the common support
    lower_limit = np.min(treated)

    # The following algorithm checks for any empty histogram bins
    # (starting from 0 going up to 0.5).
    # If an empty histogram bin is found, the lower_limit is set to
    # the corresponding P(Z) value of the next bin above.
    for low in range(len(hist[0][0])):

        # Only consider values in the interval [0, 0.5)
        if hist[1][low] > 0.5:
            break

        # If the algorithm starts below the sample minimum,
        # move on to the next bin
        elif hist[1][low] < np.min(treated):
            continue

        else:
            # If the current bin is non-empty, we have still continuous
            # support and the sample minimum remains our lower limit
            if hist[0][0][low] > 0:
                pass

            # If an empty bin is found, set the lower limit to the next bin above
            # and move on to the next bin until P(Z) = 0.5 is reached
            else:
                lower_limit = hist[1][low + 1]

    # Untreated Sample
    # Set the highest frequency observed in the untreated subsample
    # as the default for the upper limit of the common support
    upper_limit = np.max(untreated)

    # The following algorithm checks for any empty histogram bins
    # (starting from 1 going down to 0.5).
    # If an empty histogram bin is found, the upper_limit is set to the
    # current next bin.
    for up in reversed(range(len(hist[0][1]))):

        # Only consider values in the interval (0.5, 1]
        if hist[1][up] < 0.5:
            break

        # If the algorithm starts above the sample maximum, move on to the next bin
        elif hist[1][up] > np.max(untreated):
            continue

        else:
            # If the current bin is non-empty, we have still continuous support and
            # the sample maximum remains our upper limit
            if hist[0][1][up] > 0:
                pass

            # If an empty bin is found, set the upper limit to the next bin below
            # and move on to the next bin until P(Z) = 0.5 is reached
            else:
                upper_limit = hist[1][up]

    return lower_limit, upper_limit


def _construct_Xp(X, prop_score):
    """
    This function generates the X * *prop_score* regressors:

    Xp = X * P(z) for each individual.

    Parameters
    ----------
    X: pandas.DataFrame
        Data of observables (covariates).
    prop_score: pandas.Series
        Propensity score (range between [0, 1]). Values closer to 1
        denote a higher inclination to treatment.
        Sorted in ascending order.

    Returns
    -------
    X: pandas.DataFrame
        Data of observables (covariates) multiplied with *prop_score*.
    """
    # To multiply each elememt in X (shape N x k) with the corresponding prop_score,
    # set up a prop_score matrix of same size.
    P_z = pd.concat([prop_score] * len(X.columns), axis=1, ignore_index=True)

    # Construct Xp
    Xp = pd.DataFrame(
        X.values * P_z.values, columns=[key_ + "_ps" for key_ in list(X)], index=X.index
    )

    return Xp


def _generate_residuals(exog, endog, bandwidth=0.05):
    """
    This function runs a series of loess regressions (degree=1)
    for a set of response variables (*endog*) on a single explanatory
    variable (*exog*) and computes the corresponding residuals.

    For clarity, *exog* denotes Y and *endog* denotes X data.

    To avoid any confusion with the naming convention in
    the rest of this module, we use the terms *exog* and *endog* here.

    The actual loess regressions (in the terminology of the generalized
    Roy model) look like this in pseudo-code:

    For each k in NumberOfCovariates:
        X[k] = prop_score + e_X[k]
        Xp[k] = prop_score + e_Xp[k]

    And similarly:
        Y = prop_score + e_Y,

    where the goal is to retrieve e_X, e_Xp, e_Y.
    These residuals are then use by **double_residual_reg()**


    Parameters
    ----------
    exog: pandas.DataFrame or pandas.Series
        Data of exogenous variable(s). May have one- or multiple columns.
    endog: pandas.Series or pandas.DataFrame
        Single column DataFrame or Series of endogenous data.
    bandwidth: float, default is 0.05
        Bandwidth (i.e. span) used in the the loess regression.

    Returns
    -------
    res: numpy.ndarray
        Loess residuals of shape [ len(*y*), len(*y*[0]) ]
    """
    # Turn input data into np.ndarrays.
    exog = np.array(exog)
    endog = np.array(endog)

    # Determine number of observations and number of columns of the
    # outcome variable.
    n = endog.shape[0]

    # *y* is a column vector
    if endog.ndim == 1:
        y_fit = loess(exog, endog, span=bandwidth, degree=1)
        y_fit.fit()
        res = y_fit.outputs.fitted_residuals

    else:
        columns = endog.shape[1]
        res = np.zeros([n, columns])

        for col in range(columns):
            y_fit = loess(exog, endog[:, col], span=bandwidth, degree=1)
            y_fit.fit()
            res[:, col] = y_fit.outputs.fitted_residuals

    return res
